package dev.fyloz.colorrecipesexplorer.config.initializers

import dev.fyloz.colorrecipesexplorer.config.annotations.RequireDatabase
import dev.fyloz.colorrecipesexplorer.dtos.*
import dev.fyloz.colorrecipesexplorer.logic.MixLogic
import dev.fyloz.colorrecipesexplorer.utils.merge
import mu.KotlinLogging
import org.springframework.context.annotation.Configuration
import java.util.*

@Configuration
@RequireDatabase
class MixInitializer(
    private val mixLogic: MixLogic
) : AbstractInitializer() {
    private val logger = KotlinLogging.logger {}

    override fun initialize() {
        logger.debug("Executing mix initializer...")
        fixAllPositions()
    }

    private fun fixAllPositions() {
        logger.debug("Validating mix materials positions...")

        mixLogic.getAll()
            .filter { it.mixQuantities.all.any { mq -> mq.position == 0 } }
            .forEach(this::fixMixPositions)

        logger.debug("Mix materials positions are valid!")
    }

    private fun fixMixPositions(mix: MixDto) {
        val mixQuantities = mix.mixQuantitiesOutput
        val maxPosition = mixQuantities.maxOf { it.position }

        logger.warn("Mix ${mix.id} (mix name: ${mix.mixType.name}, recipe id: ${mix.recipeId}) has invalid positions:")

        val invalidMixQuantities: Collection<MixQuantityOutputDto> = with(mixQuantities.filter { it.position == 0 }) {
            if (maxPosition == 0 && this.size > 1) {
                orderMixMaterials(this)
            } else {
                this
            }
        }

        val fixedMixQuantities = increaseMixMaterialsPosition(invalidMixQuantities, maxPosition + 1)
        val updatedMixQuantities =
            mixQuantities.map { MixQuantitySaveDto(it.id, it.material.id, it.quantity, it.position, it.isMixType) }
                .merge(fixedMixQuantities)

        val updatedMix = MixSaveDto(mix.id, mix.mixType.name, mix.recipeId, mix.mixType.materialType.id, updatedMixQuantities)
        mixLogic.update(updatedMix)
    }

    private fun increaseMixMaterialsPosition(mixQuantities: Iterable<MixQuantityOutputDto>, firstPosition: Int) =
        mixQuantities
            .mapIndexed { index, mixQuantity ->
                MixQuantitySaveDto(
                    mixQuantity.id,
                    mixQuantity.material.id,
                    mixQuantity.quantity,
                    firstPosition + index,
                    mixQuantity.isMixType
                )
            }
            .onEach {
                logger.info("\tPosition of material ${it.id} (mixType: ${it.isMixType}) has been set to ${it.position}")
            }

    private fun orderMixMaterials(mixQuantities: Collection<MixQuantityOutputDto>) =
        LinkedList(mixQuantities).apply {
            while (this.peek().material.materialType.usePercentages) {
                // The first mix material can't use percents, so move it to the end of the queue
                val pop = this.pop()
                this.add(pop)
                logger.debug("\tMaterial ${pop.id} (mixType: ${pop.isMixType}) uses percents, moving to the end of the queue")
            }
        }
}