package dev.fyloz.colorrecipesexplorer.logic

import dev.fyloz.colorrecipesexplorer.dtos.EntityDto
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.exception.CannotDeleteException
import dev.fyloz.colorrecipesexplorer.exception.NotFoundException
import dev.fyloz.colorrecipesexplorer.service.Service
import dev.fyloz.colorrecipesexplorer.utils.collections.LazyMapList
import org.springframework.transaction.annotation.Transactional

/**
 * Represents the logic for a DTO type.
 *
 * @param D The type of the DTO.
 * @param S The service for the DTO.
 */
interface Logic<D : EntityDto, S : Service<D, *, *>> {
    /** Checks if a DTO with the given [id] exists. */
    fun existsById(id: Long): Boolean

    /** Get all DTOs. */
    fun getAll(): Collection<D>

    /** Get the DTO for the given [id]. Throws if no DTO were found. */
    fun getById(id: Long): D

    /** Saves the given [dto]. */
    fun save(dto: D): D

    /** Saves all the given [dtos]. */
    fun saveAll(dtos: Collection<D>): Collection<D>

    /** Updates the given [dto]. Throws if no DTO with the same id exists. */
    fun update(dto: D): D

    /** Deletes the dto with the given [id]. */
    fun deleteById(id: Long)
}

abstract class BaseLogic<D : EntityDto, S : Service<D, *, *>>(
    protected val service: S,
    protected val typeName: String
) : Logic<D, S> {
    protected val typeNameLowerCase = typeName.lowercase()

    override fun existsById(id: Long) =
        service.existsById(id)

    override fun getAll() =
        service.getAll()

    override fun getById(id: Long) =
        service.getById(id) ?: throw notFoundException(value = id)

    override fun save(dto: D) =
        service.save(dto)

    override fun saveAll(dtos: Collection<D>) =
        dtos.map(::save)

    override fun update(dto: D): D {
        if (!existsById(dto.id)) {
            throw notFoundException(value = dto.id)
        }

        return service.save(dto)
    }

    override fun deleteById(id: Long) =
        service.deleteById(id)

    protected fun notFoundException(identifierName: String = ID_IDENTIFIER_NAME, value: Any) =
        NotFoundException(
            typeNameLowerCase,
            "$typeName not found",
            "A $typeNameLowerCase with the $identifierName '$value' could not be found",
            value,
            identifierName
        )

    protected fun alreadyExistsException(identifierName: String = NAME_IDENTIFIER_NAME, value: Any) =
        AlreadyExistsException(
            typeNameLowerCase,
            "$typeName already exists",
            "A $typeNameLowerCase with the $identifierName '$value' already exists",
            value,
            identifierName
        )

    protected fun cannotDeleteException(details: String) =
        CannotDeleteException(
            typeNameLowerCase,
            "Cannot delete $typeNameLowerCase",
            details
        )

    private fun loadRelations(dto: D, relationSelectors: Collection<(D) -> Iterable<*>>) {
        relationSelectors.map { it(dto) }
            .forEach {
                if (it is LazyMapList<*, *>) {
                    it.initialize()
                } else {
                    println("Can't load :(")
                }
            }
    }

    companion object {
        const val ID_IDENTIFIER_NAME = "id"
        const val NAME_IDENTIFIER_NAME = "name"
    }
}