package dev.fyloz.colorrecipesexplorer.rest

import dev.fyloz.colorrecipesexplorer.Constants
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeEditRecipes
import dev.fyloz.colorrecipesexplorer.config.annotations.PreAuthorizeViewRecipes
import dev.fyloz.colorrecipesexplorer.dtos.RecipeDto
import dev.fyloz.colorrecipesexplorer.dtos.RecipePublicDataDto
import dev.fyloz.colorrecipesexplorer.dtos.RecipeSaveDto
import dev.fyloz.colorrecipesexplorer.dtos.RecipeUpdateDto
import dev.fyloz.colorrecipesexplorer.logic.RecipeImageLogic
import dev.fyloz.colorrecipesexplorer.logic.RecipeLogic
import org.springframework.context.annotation.Profile
import org.springframework.http.MediaType
import org.springframework.security.access.prepost.PreAuthorize
import org.springframework.web.bind.annotation.*
import org.springframework.web.multipart.MultipartFile
import javax.validation.Valid

@RestController
@RequestMapping(Constants.ControllerPaths.RECIPE)
@Profile("!emergency")
@PreAuthorizeViewRecipes
class RecipeController(private val recipeLogic: RecipeLogic, private val recipeImageLogic: RecipeImageLogic) {
    @GetMapping
    fun getAll(@RequestParam(required = false) name: String?) = ok(
        if (name == null) {
            recipeLogic.getAll()
        } else {
            recipeLogic.getAllByName(name)
        }
    )

    @GetMapping("{id}")
    fun getById(@PathVariable id: Long) = ok(recipeLogic.getById(id))

    @PostMapping
    @PreAuthorizeEditRecipes
    fun save(@Valid @RequestBody recipe: RecipeSaveDto) =
        created<RecipeDto>(Constants.ControllerPaths.RECIPE) {
            recipeLogic.save(recipe)
        }

    @PutMapping
    @PreAuthorizeEditRecipes
    fun update(@Valid @RequestBody recipe: RecipeUpdateDto) = noContent {
        recipeLogic.update(recipe)
    }

    @PutMapping("public")
    @PreAuthorize("hasAuthority('EDIT_RECIPES_PUBLIC_DATA')")
    fun updatePublicData(@Valid @RequestBody publicDataDto: RecipePublicDataDto) = noContent {
        recipeLogic.updatePublicData(publicDataDto)
    }

    @DeleteMapping("{id}")
    @PreAuthorizeEditRecipes
    fun deleteById(@PathVariable id: Long) = noContent {
        recipeLogic.deleteById(id)
    }

    @GetMapping("{recipeId}/image")
    fun getAllImages(@PathVariable recipeId: Long) =
        ok(recipeImageLogic.getAllImages(recipeId))

    @PutMapping("{recipeId}/image", consumes = [MediaType.MULTIPART_FORM_DATA_VALUE])
    @PreAuthorizeEditRecipes
    fun downloadImage(@PathVariable recipeId: Long, image: MultipartFile) =
        fileCreated("images/recipes/$recipeId") {
            recipeImageLogic.download(image, recipeId)
        }

    @DeleteMapping("{recipeId}/image/{id}")
    @PreAuthorizeEditRecipes
    fun deleteImage(@PathVariable recipeId: Long, @PathVariable id: String) = noContent {
        recipeImageLogic.delete(recipeId, id)
    }
}
