package dev.fyloz.colorrecipesexplorer.service

import dev.fyloz.colorrecipesexplorer.config.annotations.ServiceComponent
import dev.fyloz.colorrecipesexplorer.dtos.RecipeDto
import dev.fyloz.colorrecipesexplorer.dtos.RecipeGroupInformationDto
import dev.fyloz.colorrecipesexplorer.logic.config.ConfigurationLogic
import dev.fyloz.colorrecipesexplorer.model.ConfigurationType
import dev.fyloz.colorrecipesexplorer.model.Recipe
import dev.fyloz.colorrecipesexplorer.model.RecipeGroupInformation
import dev.fyloz.colorrecipesexplorer.repository.RecipeRepository
import dev.fyloz.colorrecipesexplorer.utils.collections.lazyMap
import org.springframework.transaction.annotation.Transactional
import java.time.LocalDate
import java.time.Period

interface RecipeService : Service<RecipeDto, Recipe, RecipeRepository> {
    /** Checks if a recipe with the given [name], [companyId] and a different [id] exists. */
    fun existsByNameAndCompany(name: String, companyId: Long, id: Long?): Boolean

    /** Gets all recipes with the given [name]. */
    fun getAllByName(name: String): Collection<RecipeDto>
}

@ServiceComponent
class DefaultRecipeService(
    repository: RecipeRepository,
    private val companyService: CompanyService,
    private val mixService: MixService,
    private val recipeStepService: RecipeStepService,
    private val configLogic: ConfigurationLogic
) :
    BaseService<RecipeDto, Recipe, RecipeRepository>(repository), RecipeService {
    override fun existsByNameAndCompany(name: String, companyId: Long, id: Long?) =
        repository.existsByNameAndCompanyAndIdNot(name, companyId, id ?: 0L)

    override fun getAllByName(name: String) =
        repository.findAllByName(name).map(::toDto)

    @Transactional
    override fun toDto(entity: Recipe) =
        RecipeDto(
            entity.id!!,
            entity.name,
            entity.description,
            entity.color,
            entity.gloss,
            entity.sample,
            entity.approbationDate,
            isApprobationExpired(entity) ?: false,
            entity.remark,
            companyService.toDto(entity.company),
            entity.mixes.lazyMap(mixService::toDto),
            entity.groupsInformation.lazyMap(::groupInformationToDto)
        )

    private fun groupInformationToDto(entity: RecipeGroupInformation) =
        RecipeGroupInformationDto(
            entity.id!!,
            entity.group,
            entity.note,
            entity.steps?.lazyMap(recipeStepService::toDto) ?: listOf()
        )

    override fun toEntity(dto: RecipeDto) =
        Recipe(
            dto.id,
            dto.name,
            dto.description,
            dto.color,
            dto.gloss,
            dto.sample,
            dto.approbationDate,
            dto.remark,
            companyService.toEntity(dto.company),
            dto.mixes.map(mixService::toEntity),
            dto.groupsInformation.map(::groupInformationToEntity)
        )

    private fun groupInformationToEntity(dto: RecipeGroupInformationDto) =
        RecipeGroupInformation(dto.id, dto.group, dto.note, dto.steps.map(recipeStepService::toEntity))

    private fun isApprobationExpired(recipe: Recipe): Boolean? =
        with(Period.parse(configLogic.getContent(ConfigurationType.RECIPE_APPROBATION_EXPIRATION))) {
            recipe.approbationDate?.plus(this)?.isBefore(LocalDate.now())
        }
}