package dev.fyloz.colorrecipesexplorer.logic.files

import dev.fyloz.colorrecipesexplorer.utils.FilePath
import dev.fyloz.memorycache.MemoryCache
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertNull
import kotlin.test.assertTrue

internal class DefaultFileCacheTest {
    private val memoryCacheMock = mockk<MemoryCache<String, CachedFileSystemItem>>()

    private val fileCache = spyk(DefaultFileCache(memoryCacheMock))

    private val path = FilePath("unit_test_path")
    private val cachedFile = CachedFile("unit_test_file", path, true)
    private val cachedDirectory = CachedDirectory("unit_test_dictionary", path, true)

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    private fun setup_memoryCacheMock_set() {
        every { memoryCacheMock[any()] = any() } just runs
    }

    @Test
    fun contains_normalBehavior_returnsTrue() {
        // Arrange
        every { any() in memoryCacheMock} returns true

        // Act
        val contains = path in fileCache

        // Assert
        assertTrue(contains)
    }

    @Test
    fun contains_pathNotCached_returnsFalse() {
        // Arrange
        every { any() in memoryCacheMock} returns false

        // Act
        val contains = path in fileCache

        // Assert
        assertFalse(contains)
    }

    @Test
    fun get_normalBehavior_returnsCachedItem() {
        // Arrange
        every { memoryCacheMock[any()] } returns cachedFile

        // Act
        val item = fileCache[path]

        // Assert
        assertEquals(cachedFile, item)
    }

    @Test
    fun get_pathNotCached_returnsNull() {
        // Arrange
        every { memoryCacheMock[any()] } returns null

        // Act
        val item = fileCache[path]

        // Assert
        assertNull(item)
    }

    @Test
    fun getDirectory_normalBehavior_returnsCachedDirectory() {
        // Arrange
        every { fileCache.directoryExists(any()) } returns true
        every { fileCache[any()] } returns cachedDirectory

        // Act
        val directory = fileCache.getDirectory(path)

        // Assert
        assertEquals(cachedDirectory, directory)
    }

    @Test
    fun getDirectory_directoryDoesNotExists_returnsNull() {
        // Arrange
        every { fileCache.directoryExists(any()) } returns false
        every { fileCache[any()] } returns cachedDirectory

        // Act
        val directory = fileCache.getDirectory(path)

        // Assert
        assertNull(directory)
    }

    @Test
    fun getFile_normalBehavior_returnsCachedFile() {
        // Arrange
        every { fileCache[any()] } returns cachedFile
        every { fileCache.fileExists(any()) } returns true

        // Act
        val file = fileCache.getFile(path)

        // Assert
        assertEquals(cachedFile, file)
    }

    @Test
    fun getFile_fileDoesNotExists_returnsNull() {
        // Arrange
        every { fileCache[any()] } returns cachedFile
        every { fileCache.fileExists(any()) } returns false

        // Act
        val file = fileCache.getFile(path)

        // Assert
        assertNull(file)
    }

    @Test
    fun exists_normalBehavior_returnsTrue() {
        // Arrange
        every { any() in fileCache } returns true
        every { fileCache[any()] } returns cachedFile

        // Act
        val exists = fileCache.exists(path)

        // Assert
        assertTrue(exists)
    }

    @Test
    fun exists_pathNotCached_returnsFalse() {
        // Arrange
        every { any() in fileCache } returns false
        every { fileCache[any()] } returns cachedFile

        // Act
        val exists = fileCache.exists(path)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun exists_itemDoesNotExists_returnsFalse() {
        // Arrange
        val file = cachedFile.copy(exists = false)

        every { any() in fileCache } returns true
        every { fileCache[any()] } returns file

        // Act
        val exists = fileCache.exists(path)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun directoryExists_normalBehavior_returnsTrue() {
        // Arrange
        every { fileCache.exists(any()) } returns true
        every { fileCache[any()] } returns cachedDirectory

        // Act
        val exists = fileCache.directoryExists(path)

        // Assert
        assertTrue(exists)
    }

    @Test
    fun directoryExists_pathNotCached_returnsFalse() {
        // Arrange
        every { fileCache.exists(any()) } returns false
        every { fileCache[any()] } returns cachedDirectory

        // Act
        val exists = fileCache.directoryExists(path)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun directoryExists_cachedItemIsNotDirectory_returnsFalse() {
        // Arrange
        every { fileCache.exists(any()) } returns true
        every { fileCache[any()] } returns cachedFile

        // Act
        val exists = fileCache.directoryExists(path)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun fileExists_normalBehavior_returnsTrue() {
        // Arrange
        every { fileCache.exists(any()) } returns true
        every { fileCache[any()] } returns cachedFile

        // Act
        val exists = fileCache.fileExists(path)

        // Assert
        assertTrue(exists)
    }

    @Test
    fun fileExists_pathNotCached_returnsFalse() {
        // Arrange
        every { fileCache.exists(any()) } returns false
        every { fileCache[any()] } returns cachedFile

        // Act
        val exists = fileCache.fileExists(path)

        // Assert
        assertFalse(exists)
    }

    @Test
    fun fileExists_cachedItemIsNotFile_returnsFalse() {
        // Arrange
        every { fileCache.exists(any()) } returns true
        every { fileCache[any()] } returns cachedDirectory

        // Act
        val exists = fileCache.fileExists(path)

        // Assert
        assertFalse(exists)
    }
    
    @Test
    fun setExists_normalBehavior_callsSetInCache() {
        // Arrange
        every { any() in fileCache } returns true
        every { fileCache[any()] } returns cachedFile

        setup_memoryCacheMock_set()

        val shouldExists = !cachedFile.exists

        // Act
        fileCache.setExists(path, exists = shouldExists)

        // Assert
        verify {
            memoryCacheMock[path.value] = match { it.exists == shouldExists }
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun setExists_pathNotCached_callsLoadPath() {
        // Arrange
        every { any() in fileCache } returns false
        every { fileCache[any()] } returns cachedFile

        setup_memoryCacheMock_set()

        // Act
        fileCache.setExists(path, exists = true)

        // Assert
        verify {
            fileCache.load(path)
        }
    }

    @Test
    fun load_normalBehavior_callsSetInCache() {
        // Arrange

        setup_memoryCacheMock_set()

        // Act
        fileCache.load(path)

        // Assert
        verify {
            memoryCacheMock[path.value] = match { it.path == path }
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun addItemToDirectory_normalBehavior_addsItemToDirectoryContent() {
        // Arrange
        every { fileCache.directoryExists(path) } returns true
        every { fileCache.getDirectory(path) } returns cachedDirectory

        setup_memoryCacheMock_set()

        val itemPath = FilePath("${path.value}/item")

        // Act
        fileCache.addItemToDirectory(path, itemPath)

        // Assert
        verify {
            memoryCacheMock[path.value] = match<CachedDirectory> {
                it.content.any { item -> item.path == itemPath }
            }
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun addItemToDirectory_directoryDoesNotExists_doesNothing() {
        // Arrange
        every { fileCache.directoryExists(path) } returns false
        every { fileCache.getDirectory(path) } returns cachedDirectory

        setup_memoryCacheMock_set()

        val itemPath = FilePath("${path.value}/item")

        // Act
        fileCache.addItemToDirectory(path, itemPath)

        // Assert
        verify(exactly = 0) {
            memoryCacheMock[path.value] = any()
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun addItemToDirectory_notADirectory_doesNothing() {
        // Arrange
        every { fileCache.directoryExists(path) } returns true
        every { fileCache.getDirectory(path) } returns null

        setup_memoryCacheMock_set()

        val itemPath = FilePath("${path.value}/item")

        // Act
        fileCache.addItemToDirectory(path, itemPath)

        // Assert
        verify(exactly = 0) {
            memoryCacheMock[path.value] = any()
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun removeItemFromDirectory_normalBehavior_removesItemFromDirectoryContent() {
        // Arrange
        val itemPath = FilePath("${path.value}/item")
        val file = cachedFile.copy(path = itemPath)
        val directory = cachedDirectory.copy(content = setOf(file))

        every { fileCache.directoryExists(path) } returns true
        every { fileCache.getDirectory(path) } returns directory

        setup_memoryCacheMock_set()

        // Act
        fileCache.removeItemFromDirectory(path, itemPath)

        // Assert
        verify {
            memoryCacheMock[path.value] = match<CachedDirectory> { it.content.isEmpty() }
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun removeItemFromDirectory_directoryDoesNotExists_doesNothing() {
        // Arrange
        every { fileCache.directoryExists(path) } returns false
        every { fileCache.getDirectory(path) } returns cachedDirectory

        setup_memoryCacheMock_set()

        val itemPath = FilePath("${path.value}/item")

        // Act
        fileCache.removeItemFromDirectory(path, itemPath)

        // Assert
        verify(exactly = 0) {
            memoryCacheMock[path.value] = any()
        }
        confirmVerified(memoryCacheMock)
    }

    @Test
    fun removeItemFromDirectory_notADirectory_doesNothing() {
        // Arrange
        every { fileCache.directoryExists(path) } returns true
        every { fileCache.getDirectory(path) } returns null

        setup_memoryCacheMock_set()

        val itemPath = FilePath("${path.value}/item")

        // Act
        fileCache.removeItemFromDirectory(path, itemPath)

        // Assert
        verify(exactly = 0) {
            memoryCacheMock[path.value] = any()
        }
        confirmVerified(memoryCacheMock)
    }
}